unit Big_complex;

interface
  uses system.Math, Velthuis.BigIntegers, Velthuis.Bigdecimals;
// f\
  type
    CBig = record
      r : bigdecimal;
      i : bigdecimal;
    end;

  function cadd(a, b: CBig): CBig;
  function csub(a, b: CBig): CBig;
  function cmul(a, b: Cbig): CBig;
  function cdiv(a, b: Cbig): Cbig;
  function caddb(a: cbig; b: bigdecimal): cbig;
  function csubb(a: cbig; b: bigdecimal): cbig;
  function cmulb(a: Cbig; b: bigdecimal): CBig;
  function cdivb(a: Cbig; b: bigdecimal): Cbig;
  function cabs(a : cbig; dpcs : integer): bigdecimal;
  function exp_big(x: Cbig; dpcs: integer; eps : bigdecimal): Cbig;
  function pi_big: Bigdecimal;
  function log_big(z : Cbig; eps : bigdecimal; dpcs: integer): Cbig;
  function sin_big(z : Cbig; eps : bigdecimal; dpcs: integer): Cbig;
  function cbiground(x : cbig; dpcs : integer): cbig;
  function aeqb(x, y : cbig): boolean;
  function pow_big(x, y : cbig; eps : bigdecimal): cbig;
  function cchs(x: cbig): cbig;
  function arctana_bigf(xi, pib: bigdecimal; dpcs: integer): bigdecimal;
  function arcsina_big(xi: bigdecimal; dpcs : integer): bigdecimal;
  function arg_big(z : cbig): bigdecimal;

implementation

uses main;

// cbig = -cbig
function cchs(x: cbig): cbig;
begin
  result.r := -x.r;
  result.i := -x.i;
end;

// CBigf̉Z
function cadd(a, b: CBig): CBig;
begin
  result.r := a.r + b.r;
  result.i := a.i + b.i;
end;

// CBigf̌Z
function csub(a, b: CBig): CBig;
begin
  result.r := a.r - b.r;
  result.i := a.i - b.i;
end;

// cbig + bigdecimal
function caddb(a: cbig; b:bigdecimal): cbig;
begin
  result.r := a.r + b;
  result.i := a.i;
end;

// cbig - bigdecimal
function csubb(a : cbig; b: bigdecimal): cbig;
begin
  result.r := a.r - b;
  result.i := a.i;
end;

// Cbig bigdecimal̏Z
function cmulb(a: Cbig; b: bigdecimal): CBig;
begin
  result.r := a.r * b;
  result.i := a.i * b;
end;

// Cbigf̏Z
function cmul(a, b: Cbig): CBig;
begin
  result.r := a.r * b.r - a.i * b.i;
  result.i := a.r * b.i + a.i * b.r;
  result := cbiground(result, DPS);
end;

// Cbig bigdecimal̏Z
function cdivb(a: Cbig; b: bigdecimal): Cbig;
begin
  if b <> Bigdecimal.Zero then begin
    result.r := a.r / b;
    result.i := a.i / b;
  end
  else begin
    result.r := Bigdecimal.Zero;
    result.i := Bigdecimal.Zero;
  end;
end;

// Cbigf̏Z
function cdiv(a, b: Cbig): Cbig;
var
  bb, arbraibi, aibrarbi: Bigdecimal;
begin
  bb := b.r * b.r + b.i * b.i;
  arbraibi := a.r * b.r + a.i * b.i;
  aibrarbi := a.i * b.r - a.r * b.i;
  bb := bb.RoundToPrecision(DPS);
  arbraibi := arbraibi.RoundToPrecision(DPS);
  aibrarbi := aibrarbi.RoundToPrecision(DPS);

  if bb <> Bigdecimal.Zero then begin
    result.r := arbraibi / bb;
    result.i := aibrarbi / bb;
  end
  else begin
    result.r := Bigdecimal.Zero;
    result.i := Bigdecimal.Zero;
  end;
end;

// x: cbig dpcsɊۂ
function cbiground(x : cbig; dpcs : integer): cbig;
begin
  result.r := x.r.RoundToPrecision(dpcs);
  result.i := x.i.RoundToPrecision(dpcs);
end;

// a:cbig = b:cbig
function aeqb(x, y : cbig): boolean;
begin
  if (x.r = y.r) and (x.i = y.i) then result := true
                                 else result := false;
end;

// Cbig̐Βl
function cabs(a : cbig; dpcs : integer): bigdecimal;
var
  arh2, aih2 : bigdecimal;
  tmp : bigdecimal;
begin
  arh2 := a.r * a.r;
  aih2 := a.i * a.i;
  arh2 := arh2.RoundToPrecision(dpcs);
  aih2 := aih2.RoundToPrecision(dpcs);
  tmp := arh2 + aih2;
  result := bigdecimal.Sqrt(tmp, dpcs);
end;

// exp(x)
// x ͒l f
// dpcs L
// eps l
function exp_big(x: Cbig; dpcs: integer; eps : bigdecimal): Cbig;
var
  xh, s, sb, si, ss: Cbig;
  one, ass, eight, i, ih : bigdecimal;
begin
  one := bigdecimal.One;
  eight := bigdecimal.Ten - bigdecimal.Two;
  eight := eight.RoundToPrecision(dpcs);
  x.r := x.r.RoundToPrecision(dpcs);
  x.i := x.i.RoundToPrecision(dpcs);
  x.r := x.r / eight;
  x.i := x.i / eight;
  i := one;
  ih := i;
  xh := x;
  s := x;
  repeat
    sb := s;
    i := i + one;
    ih := ih * i;             // i!
    xh := cmul(xh, x);        // x^n
    xh.r := xh.r.RoundToPrecision(dpcs);
    xh.i := xh.i.RoundToPrecision(dpcs);
    si := cdivb(xh, ih);      // x^n/i!
    s := cadd(s, si);
    ss := csub(s, sb);
    ass := cabs(ss, dpcs);
  until (ass < eps);
  s.r := s.r + one;
  s := cmul(s, s);
  s := cmul(s, s);
  result := cmul(s, s);
  result.r := result.r.RoundToPrecision(dpcs);
  result.i := result.i.RoundToPrecision(dpcs);
end;


// 
function pi_big: Bigdecimal;
var
  n, dpcs : integer;
  rmback : BigDecimal.RoundingMode;
  SQRT_SQRT_EPSILON, c, b, e : BigDecimal;
  npow : Bigdecimal;
  a, one, two, four, five, eight : Bigdecimal;
begin
  rmBack := BigDecimal.DefaultRoundingMode;           // Zۂ߃[hobNAbv
  dpcs := DPS + 5;                                    // w萸x + 
  BigDecimal.DefaultPrecision := dpcs;
  BigDecimal.DefaultRoundingMode := rmNearesteven;

  one := '1';
  one := one.RoundToPrecision(dpcs);                  // one̗Ldpcsɐݒ
  two := '2';
  four := '4';
  five := '5';
  eight := '8';
  SQRT_SQRT_EPSILON := one / BigDecimal.IntPower(two, dpcs shl 0); // l
  c := BigDecimal.Sqrt(one / eight, dpcs * 2);
  a := one + c + c + c;
  b := BigDecimal.Sqrt(a, dpcs * 2);
  e := b - five / eight;
  b := b + b;
  c := e - c;
  a := a + e;
  npow := '4';
  n := 0;
  while (e > SQRT_SQRT_EPSILON) and (n < 100) do begin
    npow := npow + npow;
    e := (a + b) / two;                               // ϒl ZLł̊ۂ
    b := a * b;
    b := b.RoundToPrecision(dpcs);                    // pcs +  ۂ
    b := BigDecimal.Sqrt(b, dpcs * 2);                // Lł̊ۂ
    e := e - b;
    b := b + b;
    c := c - e;
    a := e + b;
    inc(n);
  end;
  e := e * e;
  e := e.RoundToPrecision(dpcs);                      // pcs +  ۂ
  e := e / four;                                      // e = e * e / 4
  a := a + b;
  result := (a * a - e - e / two) / (a * c - e) / npow;   // Z̏Ԃɒ
  result := result.RoundToPrecision(DPS);             // w̐xɊۂ

  BigDecimal.DefaultPrecision := DPS;                 // ZZxA
  BigDecimal.DefaultRoundingMode := rmBack;           // Zۂ߃[hA
end;

// z ͒l f
// eps l
// dpcs L
// ndis \ true@\ false \
function log_big_sub(z : Cbig; eps : bigdecimal; dpcs: integer; ndis: boolean): Cbig;
const
  NN = 1000000;
var
  s, x, zm1, zp1, xh2, xg, n2p1 : CBig;
  one, two : Cbig;
  xgsn : Cbig;
  asabs : bigdecimal;
  n : integer;
begin
  one.r := Bigdecimal.One;      // 1
  one.i := Bigdecimal.Zero;     //
  two := cadd(one, one);        // 2
  zm1 := csub(z, one);          // z - 1
  zp1 := cadd(z, one);          // z + 1
  x := cdiv(zm1, zp1);          // x = (z-1)/(z+1)
  s.r := Bigdecimal.Zero;
  s.i := Bigdecimal.Zero;
  xh2 := cmul(x, x);            // x^2
  xg := x;                      // n = 0;    x^(2n+1)
  n2p1 := one;                  // 2n+1 = 1
  n := 0;
  repeat
    xg.r := xg.r.RoundToPrecision(dpcs);
    xg.i := xg.i.RoundToPrecision(dpcs);
    xgsn := cdiv(xg, n2p1);     // x^(2n+1) / (2n+1)
    s := cadd(s, xgsn);         // 
    xg := cmul(xg, xh2);        // x^(2n+1)
    n2p1 := cadd(n2p1, two);    // 2n + 1
    inc(n);
    asabs := cabs(xgsn, dpcs);
  until  (asabs < eps) or (n > NN);
  s := cmul(s, two);
  result := s;
end;

// z   
// eps l
// dpcs L
// Ԃl ΐf
function log_big(z : Cbig; eps : bigdecimal; dpcs: integer): Cbig;
var
  zb : Cbig;
  z10n : cbig;
  Zmax, ten, one, bn : bigdecimal;
  pi_big2: bigdecimal;
begin
  if (z.r = bigdecimal.Zero) and (z.i = bigdecimal.Zero) then begin
    result.r := bigdecimal.Zero;
    result.i := bigdecimal.Zero;
    exit;
  end;

  pi_big2 := paib / bigdecimal.Two;     // paib = pi_bigs  main.pas

  zb := z;

  // ƎƂ[łȂȂ 傫ȒlɂĉZ덷܂
  if (zb.r <> bigdecimal.Zero) and (zb.i <> bigdecimal.Zero) then begin
    if zb.i.Abs(zb.i) > zb.r.Abs(zb.r) then begin     // ̐Βl̐Βl傫ւ
      z.r := zb.i;
      z.i := zb.r;
      if zb.i < bigdecimal.Zero then begin            // 畄]
        z.r := -z.r;
        z.i := -z.i;
      end;
    end
    else begin                                        // ̐Βl̐ΒlƓ
      if zb.r < bigdecimal.Zero then begin            // ̒l畄]
        z.r := -zb.r;
        z.i := -zb.i;
      end;
    end;
  end;

  // z.r0z.i[łȂ  Ǝւ
  if (zb.r = bigdecimal.Zero) and (zb.i <> bigdecimal.Zero) then begin
    z.r := zb.i;
    z.i := zb.r;
    if z.r < bigdecimal.Zero then z.r := -z.r;      // 畄] [Ȃ̂ł̂܂
  end;

  // ̂ŋ[̎𔽓] [Ȃ̂ł̂܂
  if (zb.r <> bigdecimal.Zero) and (zb.i = bigdecimal.Zero) then
    if zb.r < bigdecimal.Zero then z.r := -zb.r;

  // 10̑ΐ𗘗pđ傫ȒľvZ𑬂܂
  ten := '10';
  ten := ten.RoundToPrecision(dpcs);               // dpcs킹
  one := bigdecimal.One;
  bn := bigdecimal.Zero;
  if not Z10F then begin
    z10.r := ten;                                       // z10 : cbig Z!0F : boolean  main.pas
    z10.i := bigdecimal.Zero;
    z10 := log_big_sub(z10, eps, dpcs, false);          // 10̑ΐ
    Z10F := true;
  end;
  zmax := cabs(z, dpcs);                              // z̐Βl
  zmax := zmax.RoundToPrecision(dpcs);             // dpcs킹
  z.r := z.r.RoundToPrecision(dpcs);               // dpcs킹
  z.i := z.i.RoundToPrecision(dpcs);               // dpcs킹
  // 10ŏZ͏ZĒl̒l1`10͈̔͂ɓ悤ɂ܂B
  repeat
    if zmax > ten then begin        // zmax10傫10ŏZ
      z.r := z.r / ten;
      z.i := z.i / ten;
      zmax := zmax / ten;
      bn := bn + one;               // 10ŏZ +ɂȂ܂B
    end;
    if zmax < one then begin        // zmax1菬10Z
      z.r := z.r * ten;
      z.i := z.i * ten;
      zmax := zmax * ten;
      bn := bn - one;               // 10Z -ɂȂ܂B
    end;
  until (zmax <= ten) and (zmax >= one);
  //----------------------------------------

  // CꂽlőΐvZ
  result := log_big_sub(z, eps, dpcs, true);

  // 10̑ΐ␳
  z10n.r := z10.r * bn;             // ZΐlvZ10̑ΐln{
  z10n.i := z10.i * bn;

  result.r := result.r + z10n.r;          // 10̑ΐln{Z
  result.i := result.i + z10n.i;
  //----------------------------------------

  // z.r[z.i[łȂ /2␳
  if (zb.r = bigdecimal.Zero) and (zb.i <> bigdecimal.Zero) then begin
    if zb.i >= bigdecimal.Zero then result.i := result.i + pi_big2
                               else result.i := result.i - pi_big2;
  end;

  // z.r̒l̂ƕύXČvZĂ̋Ε␳B
  if (zb.r < bigdecimal.Zero) and (zb.i = bigdecimal.Zero) then
    result.i := result.i + pi_big;

  // ƎƂ[łȂ
  if (zb.r <> bigdecimal.Zero) and (zb.i <> bigdecimal.Zero) then begin
    // Βl苕̐Βl傫ꍇ
    if zb.i.Abs(zb.i) > zb.r.Abs(zb.r) then begin
      if zb.i > bigdecimal.Zero then result.i := -result.i + pi_big2    // /2␳
                                else result.i := -result.i - pi_big2;
    end
    // ̎
    else begin
      if zb.r < bigdecimal.Zero then
        if zb.i > bigdecimal.Zero then result.i := result.i + pi_big    // Ε␳
                                  else result.i := result.i - pi_big
    end;
  end;
end;

// arcsin
// 0.7𒴂甽Α̊px߂܂(arccosvZɂȂ܂)
// arcsin(x) = /2 - arccos(x)
function arcsina_big(xi: bigdecimal; dpcs : integer): bigdecimal;
label
  EXT;
var
  i : integer;
  ans, bans, fa, fb, d, ab, ta, tb, f, x, seven, zero, one, two, ax: bigdecimal;
begin
  dpcs := BigDecimal.DefaultPrecision;             // ZZxobNAbv
  xi := xi.RoundToPrecision(dpcs);
  zero := bigdecimal.Zero;
  seven := '0.7';
  one := bigdecimal.One;
  two := bigdecimal.Two;
  i := 1;
  fa := one;
  x := BigDecimal.Abs(xi);
  ax := x;
  if ax = one then begin                              // 1/2 90
    ans := paib / two;
    goto EXT;
  end;
  if ax > seven then begin                            // 0.7𒴂cošvZ
    x := one - x * x;
    x := x.RoundToPrecision(dpcs);
    x := BigDecimal.Sqrt(x, dpcs * 2);                // sqrt ͗L1/2
  end;
  fb := two;
  ta := fa;
  tb := fb;
  ab := '3';
  d := x * x * x;
  d := d.RoundToPrecision(dpcs);
  ans := (x + d * (ta / tb / ab));
  repeat
    bans := ans;
    d := d * x * x;
    d := d.RoundToPrecision(dpcs);
    fa := fa + two;
    fb := fb + two;
    ab := ab + two;
    ta := ta * fa;
    ta := ta.RoundToPrecision(dpcs);
    tb := tb * fb;
    tb := tb.RoundToPrecision(dpcs);
    f := (ta / tb / ab) * d;
    ans := ans + f;
    inc(i);
  until (bans = ans) or (i > 10000);
  if ax > seven then ans := paib / two - ans;
EXT:
  ans := ans.RoundToPrecision(dpcs);
  if xi < zero then ans := -ans;
  result := ans;
end;

// sin(z)
// eps l
// dpcs L
// Ԃl sinf
function sin_big(z : Cbig; eps : bigdecimal; dpcs: integer): Cbig;
var
  zh2n1, zh2: cbig;
  fa2n1, fn, one : bigdecimal;
  s, sb, tmp : cbig;
  abss: bigdecimal;
  FG : boolean;
begin
  // n = 0
  one := bigdecimal.One;
  zh2n1 := z;                     // z^(2n+1)
  zh2 := cmul(z,z);               // z^2
  fa2n1 := one;                   // (2n+1)!
  fn := one;                      // (2*0+1)!
  s := z;                         // z
  FG := false;                    // ͊
  // n = 1 `
  repeat
    sb := s;
    zh2n1 := cmul(zh2n1, zh2);    // z^(2n+1) = z^(2n+1) * Z^2
    fn := fn + one;               // fn = fn + 1
    fa2n1 := fa2n1* fn;           // (2n+1)! = fa2n1*fn
    fn := fn + one;               // fn = fn + 1
    fa2n1 := fa2n1* fn;           // (2n+1)!  = fa2n1*fn
    // ZOɌ
    fa2n1 := fa2n1.RoundToPrecision(dpcs);
    zh2n1.r := zh2n1.r.RoundToPrecision(dpcs);
    zh2n1.i := zh2n1.i.RoundToPrecision(dpcs);
    tmp := cdivb(zh2n1, fa2n1);   // z^(2n+1) / (2n+1)!
    if FG then begin              // Ȃ
      s := cadd(s, tmp);          // Z
      FG := false;                // ͊
    end
    else begin                    // Ȃ
      s := csub(s, tmp);          // Z
      FG := true;                 // ͋
    end;
    abss := cabs(tmp, dpcs);
  until abss < eps;
  result := s;
end;

// x^y
// exp[ln(x) * y]
// A[ɂȂłAZ덷ɂA[ɂȂȂ̂āAċI
// Ƀ[ɂĂ܂B
function pow_big(x, y : cbig; eps : bigdecimal): cbig;
var
  dpcs : integer;
  harf, absi, absy, four : bigdecimal;
  lnx, tmp : cbig;
begin
  if (x.r = bigdecimal.zero) and (x.i = bigdecimal.zero) then begin
    result.r := bigdecimal.Zero;
    result.i := bigdecimal.Zero;
    exit;
  end;
  dpcs := BigDecimal.DefaultPrecision;
  lnx := log_big(x, eps, dpcs);               // ln(x)
  lnx := cbiground(lnx, dpcs);                // ppcs ۂ
  y := cbiground(y, dpcs);                    // ppcs ۂ
  tmp := cmul(lnx, y);
  tmp := cbiground(tmp, dpcs);                // ln(x) * y
  result := exp_big(tmp, dpcs, eps);

  // x̋y̋0Ȃ
  if (x.i = bigdecimal.zero) and (y.i = bigdecimal.Zero) then begin
    harf := bigdecimal.One / bigdecimal.Two;          // 0.5
    absy := y.r.Abs(y.r);       // y̎̐Βl
    absi := absy.Frac;          // y̎̐Βl̏
    absi := absi - harf;        // y̎̐Βl̏0.5̍
    // [(y̎̏0.5)
    if absi = bigdecimal.Zero then
      // x̎0Ɠ0傫Ȃ
      if x.r >= bigdecimal.zero then result.i := bigdecimal.zero   // ̋0
                                // 0菬Ȃ
                                else result.r := bigdecimal.zero;  // ̎0
    absi := y.r.Frac;           // 搔̏
    if absi = bigdecimal.Zero then result.i := bigdecimal.Zero;    // Ȃ瓚̋0
  end;
  // x̐y̋0Ȃ
  if (x.r = bigdecimal.Zero) and (y.i = bigdecimal.Zero) then begin
    absi := y.r.Frac;                                     // 搔̏
    if absi = bigdecimal.Zero then begin                  // Ȃ
      absy := y.r / bigdecimal.Two;                       // mF
      absi := absy.Frac;                                  // 
      if absi = bigdecimal.Zero then result.i := bigdecimal.Zero             // Ȃ狕0
                                else result.r := bigdecimal.Zero;            // Ȃ0
    end;
  end;
  absy := x.r.Abs(x.r);                                     // |x.r|
  absi := x.i.Abs(x.i);                                     // |x.i|
  // x̎Ƌ̐Βly.r[łȂ
  if (absy = absi) and (y.i = bigdecimal.Zero) and not (y.r = bigdecimal.Zero) then begin
    absi := y.r.Frac;                                       // 搔̏
    if absi = bigdecimal.Zero then  begin                   // Ȃ
      absy := y.r / bigdecimal.Two;                         // mF
      absi := absy.Frac;                                    // 
      if absi = bigdecimal.Zero then begin                  // Ȃ
        four := bigdecimal.Two + bigdecimal.Two;            // 4
        absy := y.r / four;                                 // 4mF
        absi := absy.Frac;                                  // 
        if absi = bigdecimal.Zero then result.i := bigdecimal.Zero      // Ȃ 0  y= 4 8
                                  else result.r := bigdecimal.Zero;     // Ȃ 0  y= 2 6 10
      end;
    end;
  end;
end;


// arctan
// 1 z甽Α̊pxvZ
// IngEIC[̌vZ
function arctana_bigf(xi, pib: bigdecimal; dpcs: integer): bigdecimal;
var
  i : integer;
  x, ax, y, ans, bans, f, ib, zero, one, two, a2: bigdecimal;
begin
  xi := xi.RoundToPrecision(dpcs);
  x := bigdecimal.Abs(xi);
  ax := x;
  zero := '0';
  one := '1';
  two := '2';
  if x > one then ax := one / x;
  a2 := ax * ax;
  a2 := a2.RoundToPrecision(dpcs);
  y := a2 / (one + a2);
//  y := y.RoundToScale(dpcs, rmNearesteven);
  i := 0;
  ib := one;
  ans := one;
  f := one;
  repeat
    bans := ans;
    f := f * ib * (two / (ib * two + one)) * y;
    f := f.RoundToPrecision(dpcs);
    ans := ans  + f;
    ans := ans.RoundToPrecision(dpcs);
    ib := ib + one;
    inc(i);
  until (bans = ans) or (i > 10000);
  if ax <> zero then ans := ans * (y / ax)
                else ans := zero;
  if x > one then ans := pib / two - ans;
  if xi < zero then ans := -ans;

  result := ans;
end;

// fΊpvZ
function arg_big(z : cbig): bigdecimal;
var
  zin : cbig;
  x, y, ysx, pi1s2, zero : bigdecimal;
  dpcs :integer;
begin
  dpcs := BigDecimal.DefaultPrecision;
  zero := bigdecimal.Zero;
  zin := cbiground(z, dpcs);
  result := zero;
  x := zin.r;
  y := zin.i;
  if (x = zero) and (y = zero) then exit;
  pi1s2 := paib / bigdecimal.two;
  if (x = zero) and (y > zero) then result := pi1s2;
  if (x = zero) and (y < zero) then result := - pi1s2;

  if x <> zero then begin
    ysx := y / x;
    if x > zero then result := arctana_bigf(ysx, paib, dpcs);
    if x < zero then begin
      if y >= zero then result := arctana_bigf(ysx, paib, dpcs) + paib;
      if y < zero then result := arctana_bigf(ysx, paib, dpcs) - paib;
    end;
  end;
end;

end.
